#include <tuple>
#include <utility>

#if __cplusplus > 202002L

template <typename... Ts>
constexpr int sum(std::tuple<Ts...> vs) {
	return [&]<std::size_t... Is>(std::index_sequence<Is...>) {
		return (std::get<Is>(vs) + ...);
	}
	(std::index_sequence_for<Ts...>{});
}

int basic_sum() {
	// return sum(std::tuple{0, 1, 2, 3});
	int i{};
	return sum(std::tuple{i++, i++, i++, i++});
}

template <typename... Ts>
struct curry {
	using tuple_t = std::tuple<Ts...>;
	tuple_t vs;

	curry(Ts&&... ts) : vs(std::forward<Ts>(ts)...) {}
	curry(tuple_t&& t) : vs(std::move(t)) {}

	template <typename U>
	constexpr auto operator()(U&& u) const noexcept -> curry<Ts..., U> {
		return {std::tuple_cat(vs, std::forward_as_tuple(std::forward<U>(u)))};
	}

	template <typename F>
	constexpr auto apply(F&& f) -> decltype(auto) {
		return std::apply(f, vs);
	}
};

template <std::size_t I, typename... Ts>
auto get(const curry<Ts...>& b)
    -> std::tuple_element_t<I, const std::tuple<Ts...>>& {
	return std::get<I>(b.vs);
}

template <typename... Ts>
constexpr int add(curry<Ts...> vs) {
	return [&]<std::size_t... Is>(std::index_sequence<Is...>) {
		return (get<Is>(vs) + ...);
	}
	(std::index_sequence_for<Ts...>{});
}

int build_sum() {
	// return add(build(1)(2)(3));
	int i{};
	return add(curry(i++)(i++)(i++)(i++));
}

template <typename T, template <typename...> typename Class>
struct is_instantiation_of_impl : std::false_type {};
template <typename... Ts, template <typename...> typename Class>
struct is_instantiation_of_impl<Class<Ts...>, Class> : std::true_type {};

template <typename T, template <typename...> typename Class>
concept is_instantiation_of = is_instantiation_of_impl<T, Class>::value;

template <typename... Ts>
constexpr auto add(Ts... ts) requires(sizeof...(Ts) > 1 or
                                      (not is_instantiation_of<Ts, curry> &&
                                       ...)) {
	return (ts + ... + 0);
}

int apply_sum() {
	int i{};
	return curry(i++)(i++)(i++)(i++).apply(
	    [](auto... ts) { return add(ts...); });
}

#endif
